{
  "type": "record",
  "name": "ChartQuery",
  "namespace": "com.linkedin.pegasus2avro.chart",
  "doc": "Information for chart query which is used for getting data of the chart",
  "fields": [
    {
      "name": "rawQuery",
      "type": "string",
      "doc": "Raw query to build a chart from input datasets"
    },
    {
      "name": "type",
      "type": {
        "type": "enum",
        "name": "ChartQueryType",
        "symbols": [
          "LOOKML",
          "SQL"
        ],
        "symbolDocs": {
          "LOOKML": "LookML queries",
          "SQL": "SQL type queries"
        }
      },
      "doc": "Chart query type",
      "Searchable": {
        "addToFilters": true,
        "fieldName": "queryType",
        "fieldType": "KEYWORD",
        "filterNameOverride": "Query Type"
      }
    }
  ],
  "Aspect": {
    "name": "chartQuery"
  }
}